unit AVI;

interface

uses
  Windows, Graphics, DIBitmap, ActiveX, SysUtils;

type
  EAVIError = class (Exception);

  LONG = Longint;
  PVOID = Pointer;

  TAVIFileInfoW = record
    dwMaxBytesPerSec,	// max. transfer rate
    dwFlags,		// the ever-present flags
    dwCaps,
    dwStreams,
    dwSuggestedBufferSize,

    dwWidth,
    dwHeight,

    dwScale,
    dwRate,	// dwRate / dwScale == samples/second
    dwLength,

    dwEditCount: DWORD;

    szFileType: array[0..63] of WideChar;		// descriptive string for file type?
  end;
  PAVIFileInfoW = ^TAVIFileInfoW;

  TAVIStreamInfoA = record
    fccType,
    fccHandler,
    dwFlags,        // Contains AVITF_* flags
    dwCaps: DWORD;
    wPriority,
    wLanguage: WORD;
    dwScale,
    dwRate, // dwRate / dwScale == samples/second
    dwStart,
    dwLength, // In units above...
    dwInitialFrames,
    dwSuggestedBufferSize,
    dwQuality,
    dwSampleSize: DWORD;
    rcFrame: TRect;
    dwEditCount,
    dwFormatChangeCount,
    szName:  array[0..63] of AnsiChar;
  end;
  TAVIStreamInfo = TAVIStreamInfoA;

  TAVIStreamInfoW = record
    fccType,
    fccHandler,
    dwFlags,        // Contains AVITF_* flags
    dwCaps: DWORD;
    wPriority,
    wLanguage: WORD;
    dwScale,
    dwRate, // dwRate / dwScale == samples/second
    dwStart,
    dwLength, // In units above...
    dwInitialFrames,
    dwSuggestedBufferSize,
    dwQuality,
    dwSampleSize: DWORD;
    rcFrame: TRect;
    dwEditCount,
    dwFormatChangeCount,
    szName:  array[0..63] of WideChar;
  end;

  IAVIStream = interface (IUnknown)
    ['{00020021-0000-0000-C000-000000000046}']
    function Create(lParam1, lParam2: LPARAM): HResult; stdcall;
    function Info(var psi: TAVIStreamInfoW; lSize: LONG): HResult; stdcall;
    function FindSample(lPos, lFlags: LONG): LONG; stdcall;
    function ReadFormat(lPos: LONG; lpFormat: PVOID; var lpcbFormat: LONG): HResult; stdcall;
    function SetFormat(lPos: LONG; lpFormat: PVOID; lpcbFormat: LONG): HResult; stdcall;
    function Read(lStart, lSamples: LONG; lpBuffer: PVOID; cbBuffer: LONG; var plBytes: LONG; var plSamples: LONG): HResult; stdcall;
    function Write(lStart, lSamples: LONG; lpBuffer: PVOID; cbBuffer: LONG; dwFlags: DWORD; var plSampWritten: LONG; var plBytesWritten: LONG): HResult; stdcall;
    function Delete(lStart, lSamples: LONG): HResult; stdcall;
    function ReadData(fcc: DWORD; lp: PVOID; var lpcb: LONG): HResult; stdcall;
    function WriteData(fcc: DWORD; lp: PVOID; cb:  LONG): HResult; stdcall;
    function SetInfo(var lpInfo: TAVIStreamInfoW; cbInfo: LONG): HResult; stdcall;
  end;

  IAVIFile = interface (IUnknown)
    ['{00020020-0000-0000-C000-000000000046}']
    function Info(var pfi: TAVIFileInfoW; lSize: LONG): HResult; stdcall;
    function GetStream(var ppStream: IAVIStream; fccType: DWORD; lParam: LONG): HResult; stdcall;
    function CreateStream(var ppStream: IAVIStream; var pfi: TAVIFileInfoW): HResult; stdcall;
    function WriteData(ckid: DWORD; lpData: PVOID; cbData: LONG): HResult; stdcall;
    function ReadData(ckid: DWORD; lpData: PVOID; var lpcbData: LONG): HResult; stdcall;
    function EndRecord: HResult; stdcall;
    function DeleteStream(fccType: DWORD; lParam: LONG): HResult; stdcall;
  end;

  TAVIStream = class (TObject)
  private
    FFile : IAVIFile;
    FStream : IAVIStream;
    FStreamInfo : TAVIStreamInfo;
    FCount : Integer;
    FRate : Integer;

    FBitmapInfo: PBitmapInfoHeader;
    FBitmapInfoSize: Integer;
    FBitmapBits: Pointer;
    FBitmapSize: Cardinal;
  public
    constructor Create(FileName : String);
    destructor Destroy; override;
    procedure AddBitmap(B : TBitmap);
  published
    property Rate : Integer read FRate write FRate;
  end;

  procedure AVIFileInit; stdcall;
  procedure AVIFileExit; stdcall;
  function AVIFileOpen(var ppfile: IAVIFile; szFile: LPCSTR; uMode: UINT; lpHandler: PCLSID): HResult; stdcall;
  function AVIFileCreateStream(pfile: IAVIFile; var ppavi: IAVISTREAM; var psi: TAVIStreamInfoA): HResult; stdcall;
  function AVIStreamSetFormat(pavi: IAVIStream; lPos: LONG; lpFormat: PVOID; cbFormat: LONG): HResult; stdcall;
  function AVIStreamWrite(pavi: IAVIStream; lStart, lSamples: LONG; lpBuffer: PVOID; cbBuffer: LONG; dwFlags: DWORD; var plSampWritten: LONG; var plBytesWritten: LONG): HResult; stdcall;
  function AVIStreamRelease(pavi: IAVISTREAM): ULONG; stdcall;
  function AVIFileRelease(pfile: IAVIFile): ULONG; stdcall;

implementation

type
  AVIErrors = (avierrUnsupported, avierrBadFormat, avierrMemory, avierrInternal,
               avierrBadFlags, avierrBadParam, avierrBadSize, avierrBadHandle,
               avierrFileRead, avierrFileWrite, avierrFileOpen, avierrCompressor,
               avierrNoCompressor, avierrReadOnly, avierrNoData, avierrBufferTooSmall,
               avierrCantCompress);

const
  FACILITY_ITF = $4;
  SEVERITY_ERROR = $1;

  AVIERR_OK = 0;
  AVIErrorStart = (SEVERITY_ERROR shl 31) or (FACILITY_ITF shl 16) or ($4000 + 101);

  AVIERR_USERABORT      = (SEVERITY_ERROR shl 31) or (FACILITY_ITF shl 16) or ($4000 + 198);
  AVIERR_ERROR          = (SEVERITY_ERROR shl 31) or (FACILITY_ITF shl 16) or ($4000 + 199);

  AVIErrorMsg : Array[AVIErrors] of String = (
    'Unsupported',
    'Corrupt file or an unrecognized format',
    'Insufficient memory',
    'Internal',
    'Bad Flags',
    'Bad Param',
    'Bad Size',
    'Bad Handle',
    'Cannot read file',
    'Cannot write file',
    'Cannot open file',
    'Compressor',
    'No Compressor',
    'Read Only',
    'No Data',
    'Buffer Too Small',
    'Can''t Compress');


   // AVI Interface IDs
  IID_IAVIStreaming  : TGUID = '{00020022-0000-0000-C000-000000000046}';
  IID_IGetFrame      : TGUID = '{00020023-0000-0000-C000-000000000046}';
  IID_IAVIEditStream : TGUID = '{00020024-0000-0000-C000-000000000046}';

   // AVI Class IDs
  CLSID_AVISimpleUnMarshal : TGUID = '{00020009-0000-0000-C000-000000000046}';
  CLSID_AVIFile            : TGUID = '{00020000-0000-0000-C000-000000000046}';

  AVIIF_LIST      = $01;
  AVIIF_TWOCC	  = $02;
  AVIIF_KEYFRAME  = $10;

  streamtypeVIDEO = $73646976; // DWORD( 'v', 'i', 'd', 's' )

procedure AVIFileInit; stdcall; external 'avifil32.dll' name 'AVIFileInit';
procedure AVIFileExit; stdcall; external 'avifil32.dll' name 'AVIFileExit';
function AVIFileOpen(var ppfile: IAVIFILE; szFile: LPCSTR; uMode: UINT; lpHandler: PCLSID): HResult; external 'avifil32.dll' name 'AVIFileOpenA';
function AVIFileCreateStream(pfile: IAVIFile; var ppavi: IAVIStream; var psi: TAVIStreamInfoA): HResult; external 'avifil32.dll' name 'AVIFileCreateStreamA';
function AVIStreamSetFormat(pavi: IAVIStream; lPos: LONG; lpFormat: PVOID; cbFormat: LONG): HResult; external 'avifil32.dll' name 'AVIStreamSetFormat';
function AVIStreamWrite(pavi: IAVIStream; lStart, lSamples: LONG; lpBuffer: PVOID; cbBuffer: LONG; dwFlags: DWORD; var plSampWritten: LONG; var plBytesWritten: LONG): HResult; external 'avifil32.dll' name 'AVIStreamWrite';
function AVIStreamRelease(pavi: IAVIStream): ULONG; external 'avifil32.dll' name 'AVIStreamRelease';
function AVIFileRelease(pfile: IAVIFile): ULONG; external 'avifil32.dll' name 'AVIFileRelease';

procedure AVICheck(Value : HResult);
begin
  if Value = AVIERR_OK then Exit;

  if Value - AVIErrorStart <= Ord(High(AVIErrors)) then
    raise EAVIError.Create('AVI Error: ' + AVIErrorMsg[AVIErrors(Value - AVIErrorStart)])
  else
    raise Exception.Create(SysErrorMessage(Value));
end;

constructor TAVIStream.Create(FileName : String);
begin
  inherited Create;

  FRate := 10;

  AVIFileInit;
  AVICheck(AVIFileOpen(FFile, PChar(FileName), OF_WRITE or OF_CREATE, nil));
end;

destructor TAVIStream.Destroy;
begin
  if FBitmapInfo <> Nil then FreeMem(FBitmapInfo);
  if FBitmapBits <> Nil then FreeMem(FBitmapBits);
  AVIFileExit;
  inherited Destroy;
end;

procedure TAVIStream.AddBitmap(B : TBitmap);
var
  nul: Longint;
begin
  if FStream = Nil then
    begin
       // Base the AVI off the first image in the list
      with B do
        begin
          InternalGetDIBSizes(Handle,FBitmapInfoSize,FBitmapSize,PixelFormat);
          FBitmapInfo := AllocMem(FBitmapInfoSize);
          FBitmapBits := AllocMem(FBitmapSize);
          InternalGetDIB(Handle,0,FBitmapInfo^,FBitmapBits^,PixelFormat);
        end;

      with FStreamInfo do
        begin
          fccType := streamtypeVIDEO;
          fccHandler := 0;
          dwScale := 1;
          dwRate := FRate;

          dwSuggestedBufferSize := FBitmapInfo^.biSizeImage;
          rcFrame.Right := FBitmapInfo^.biWidth;
          rcFrame.Bottom := FBitmapInfo^.biHeight;
        end;

      AVICheck(AVIFileCreateStream(FFile, FStream, FStreamInfo));
      InternalGetDIB(B.Handle,0,FBitmapInfo^,FBitmapBits^,B.PixelFormat);
      AVICheck(AVIStreamSetFormat(FStream,0,FBitmapInfo,FBitmapInfoSize));
    end;

  with B do
    begin
      InternalGetDIB(Handle,0,FBitmapInfo^,FBitmapBits^,PixelFormat);
      AVICheck(AVIStreamWrite(FStream,FCount,1,FBitmapBits,FBitmapSize,AVIIF_KEYFRAME,nul,nul));
      Inc(FCount);
    end;
end;

end.
