unit MemMapFile;

interface

uses
  Windows, Classes;

type
  TMMFStream = class (TStream)
  private
    FileHandle : THandle;
    MapHandle : THandle;
    FMemory : Pointer;
    FSize : Integer;
    FOffset : Integer;
  public
    constructor Create(FileName : String);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    property Memory : Pointer read FMemory;
  end;

implementation

uses
  SysUtils, Util;

constructor TMMFStream.Create(FileName : String);
var
  I : Integer;
begin
  inherited Create;

   // Open the file
  FileHandle := CreateFile(
         PChar(FileName),               // File name
         GENERIC_READ,        // Access (0, GENERIC_READ, GENERIC_WRITE)
         FILE_SHARE_READ,     // Sharing (0, FILE_SHARE_READ, FILE_SHARE_WRITE)
         Nil,           // Security settings
         OPEN_EXISTING,       // How to create
         FILE_FLAG_RANDOM_ACCESS,  // Flags and attributes
         0); // handle of file with attributes to copy
  if FileHandle = INVALID_HANDLE_VALUE then Win32CheckKnown(GetLastError);
  try
    FSize := GetFileSize(FileHandle, @I);

     // Create the mapping
    MapHandle := CreateFileMapping(
            FileHandle,
            Nil,
            PAGE_READONLY,
            0, FSize,
            NIL);
    if MapHandle = 0 then Win32CheckKnown(GetLastError);
    try
       // Map it!
      FMemory := MapViewOfFile(  // Ex?
             MapHandle,
             FILE_MAP_READ,
             0, 0,
             FSize);
      if FMemory = Nil then raise Exception.Create('Could not map a view of the file');
    except
      CloseHandle(MapHandle);
      MapHandle := 0;
      raise;
    end;
  except
    CloseHandle(FileHandle);
    FileHandle := INVALID_HANDLE_VALUE;
    raise;
  end;
end;

destructor TMMFStream.Destroy;
begin
  if FMemory <> Nil then UnmapViewOfFile(FMemory);
  if MapHandle <> 0 then CloseHandle(MapHandle);
  if FileHandle <> INVALID_HANDLE_VALUE then CloseHandle(FileHandle);
  inherited Destroy;
end;

function TMMFStream.Read(var Buffer; Count: Longint): Longint;
begin
  if FOffset + Count > FSize then
    Count := FSize - FOffset;
  Move(PChar(FMemory)[FOffset], Buffer, Count);
  Result := Count;
end;

function TMMFStream.Write(const Buffer; Count: Longint): Longint;
begin
  raise Exception.Create('Cannot write to TMMFStream');
end;

function TMMFStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
    soFromBeginning : FOffset := Offset;
    soFromCurrent : FOffset := FOffset + Offset;
    soFromEnd : FOffset := FSize - Offset;
  end;
  Result := FOffset;
end;

end.
