unit Version;

interface

uses
  Windows;

type
  TVersionInformation = record
    FileVersionMS,
    FileVersionLS,
    ProductVersionMS,
    ProductVersionLS : DWord;
    CompanyName,
    FileDescription,
    FileVersion,
    InternalName,
    LegalCopyright,
    OriginalFileName,
    ProductName,
    ProductVersion : String;
  end;

function DDWordToStr(D1, D2 : DWord) : String;
procedure GetVersionInformation(var VersionInformation : TVersionInformation;
                                Language, SubLanguage, CodePage : Word);

 // See StringTable for explanation of Language, SubLanguage and CodePage.
 // Also, reference Windows.pas for the constant values.

implementation

uses
  Resource, SysUtils, Classes;

type
  TChildren = (chStringFileInfo, chVarFileInfo);

function DDWordToStr(D1, D2 : DWord) : String;
begin
  Result := IntToHex(HiWord(D1),2)+'.'+
            IntToHex(LoWord(D1),2)+'.'+
            IntToHex(HiWord(D2),2)+'.'+
            IntToHex(LoWord(D2),2);
end;

procedure GetVersionInformation(var VersionInformation : TVersionInformation;
                                Language, SubLanguage, CodePage : Word);

var
  LanguageStr : String;

  procedure ProcessTranslation(S : TExtResSubStream; ValA, ValB : Word);
  //var
  //  BSize : Integer;
  begin
  //  BSize := S.Size-S.Position;
    // If you so desire, you may read in an array of DWords that will indicate
    // the languages supported by this exe at this point.
  end;

  procedure ProcessVarFileInfo(S : TExtResSubStream; ValA, ValB : Word);
   // VarFileInfo
  var
    wLength, wValA, wValB : Word;
    SubStream : TExtResSubStream;
    Name : String;
    P : Integer;
  begin
    While S.Position < S.Size do
      begin
        P := S.Position;
        S.Read(wLength, SizeOf(wLength));
        SubStream := TExtResSubStream.Create;
        try
          with SubStream do
            begin
              SetPointer(PChar(S.Memory)+P,wLength);
              Read(wLength, SizeOf(wLength));
              Read(wValA, SizeOf(wValA));
              Read(wValB, SizeOf(wValB));
              Name := LowerCase(ReadKey);
            end;

          If Name = 'translation' then
            ProcessTranslation(SubStream,wValA,wValB);
        finally
          SubStream.Free;
        end;

        S.Seek(P+wLength,soFromBeginning);
      end;
  end;

  procedure ProcessStrings(S : TExtResSubStream; ValA, ValB : Word);
   // Strings
  type
    CharArray = Array[0..65519] of Char;
    PCharArray = ^CharArray;
  var
    wLength, // Does not include padding between strings
    wCharCount,
    wValA : Word;
    Name : String;
    Value : String;
  begin
    While S.Position < S.Size do
      begin
        S.Read(wLength, SizeOf(wLength));
        S.Read(wCharCount, SizeOf(wCharCount));
        S.Read(wValA, SizeOf(wValA));
        Name := LowerCase(S.ReadKey);
        Value := S.ReadKey;
        If      Name = 'companyname'      then VersionInformation.CompanyName := Value
        else if Name = 'filedescription'  then VersionInformation.FileDescription := Value
        else if Name = 'fileversion'      then VersionInformation.FileVersion := Value
        else if Name = 'internalname'     then VersionInformation.InternalName := Value
        else if Name = 'legalcopyright'   then VersionInformation.LegalCopyright := Value
        else if Name = 'originalfilename' then VersionInformation.OriginalFileName := Value
        else if Name = 'productname'      then VersionInformation.ProductName := Value
        else if Name = 'productversion'   then VersionInformation.ProductVersion := Value;
      end;
  end;

  procedure ProcessStringFileInfo(S : TExtResSubStream; ValA, ValB : Word);
   // StringFileInfo
  var
    wLength, wValA, wValB : Word;
    SubStream : TExtResSubStream;
    Name : String;
    P : Integer;
  begin
    While S.Position < S.Size do
      begin
        P := S.Position;
        S.Read(wLength, SizeOf(wLength));
        SubStream := TExtResSubStream.Create;
        try
          with SubStream do
            begin
              SetPointer(PChar(S.Memory)+P,wLength);
              Read(wLength, SizeOf(wLength));
              Read(wValA, SizeOf(wValA));
              Read(wValB, SizeOf(wValB));
              Name := LowerCase(ReadKey);
            end;

          If Name = LanguageStr then
            ProcessStrings(SubStream,wValA,wValB);
        finally
          SubStream.Free;
        end;

        S.Seek(P+wLength,soFromBeginning);
      end;
  end;

  procedure ProcessVersionInfo(S : TExtResSubStream; ValA, ValB : Word);
  var
    wLength, wValA, wValB : Word;
    Name : String;
    Value : TVSFixedFileInfo;
    SubStream : TExtResSubStream;
    P : Integer;
  begin
    FillChar(Value,SizeOf(Value),0);

     // Load VS_VERSION_INFO.Value
    S.Read(Value,ValA);
    If ValA >= 16 then
      begin
        VersionInformation.FileVersionMS := Value.dwFileVersionMS;
        VersionInformation.FileVersionLS := Value.dwFileVersionLS;
      end;
    If ValA >= 24 then
      begin
        VersionInformation.ProductVersionMS := Value.dwProductVersionMS;
        VersionInformation.ProductVersionLS := Value.dwProductVersionLS;
      end;

     // Load VS_VERSION_INFO.Padding2
    S.ReadPadding;

    While S.Position < S.Size do
      begin
        P := S.Position;
        S.Read(wLength, SizeOf(wLength));
        SubStream := TExtResSubStream.Create;
        try
          with SubStream do
            begin
              SetPointer(PChar(S.Memory)+P,wLength);
              Read(wLength, SizeOf(wLength));
              Read(wValA, SizeOf(wValA));
              Read(wValB, SizeOf(wValB));
              Name := LowerCase(ReadKey);
            end;

          If Name = 'varfileinfo' then
            ProcessVarFileInfo(SubStream,wValA,wValB)
          Else if Name = 'stringfileinfo' then
            ProcessStringFileInfo(SubStream,wValA,wValB);
        finally
          SubStream.Free;
        end;

        if wLength <> 0 then
          S.Seek(P+wLength,soFromBeginning);
      end;
  end;

  procedure ProcessStream(S : TExtResStream);
  var
    wLength, wValA, wValB : Word;
    SubStream : TExtResSubStream;
    P : Integer;
    Name : String;
  begin
    P := S.Position;
    S.Read(wLength, SizeOf(wLength));
    SubStream := TExtResSubStream.Create;
    try
      with SubStream do
        begin
          SetPointer(PChar(S.Memory)+P,wLength);
          Read(wLength, SizeOf(wLength));
          Read(wValA, SizeOf(wValA));
          Read(wValB, SizeOf(wValB));
          Name := ReadKey;
        end;

      If Name = 'VS_VERSION_INFO' then
        ProcessVersionInfo(SubStream, wValA, wValB);
    finally
      SubStream.Free;
    end;
  end;

var
  ResourceStream : TExtResStream;

begin
  LanguageStr := LowerCase(
                   IntToHex((SubLanguage shl 10) + Language,4) +
                   IntToHex(CodePage,4));
  ResourceStream := TExtResStream.CreateFromID(HInstance, 1, RT_VERSION);
  try
    ProcessStream(ResourceStream);
  finally
    ResourceStream.Free;
  end;
end;

end.
