unit ArtView;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, ActnList, ImgList, ComCtrls, ToolWin;

type
  TfrmArtView = class(TForm)
    pnlHeader: TPanel;
    pnlOuter: TPanel;
    imgTile: TImage;
    pnlHeaderText: TPanel;
    lblBlockNum: TLabel;
    lblBlockNumTitle: TLabel;
    ToolBar: TToolBar;
    tbSave: TToolButton;
    ilButtons: TImageList;
    alButtons: TActionList;
    actSave: TAction;
    SaveDialog: TSaveDialog;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure actSaveExecute(Sender: TObject);
    procedure FormResize(Sender: TObject);
  public
    procedure ShowBitmap(
      AName : String; ABitmap : TBitmap;
      AIndexTitle : String;
      AIndex : Integer; AAppendIndex : Boolean);
  private
    Bitmap : TBitmap;
  end;

implementation

{$R *.DFM}

uses
  Utility;

function StripDamagingChars(S : String) : String;
var
  I : Integer;
  Prev : Char;
begin
  Prev := ' ';
  for I := 1 to Length(S) do
    begin
      case S[I] of
        'A'..'Z', 'a'..'z', '0'..'9', '_', '-' :
          begin
            Result := Result + S[I];
            Prev := S[I];
          end;
        ' ' :
          if Prev <> ' ' then
            begin
              Result := Result + ' ';
              Prev := ' ';
            end;
      end;
    end;
end;

procedure TfrmArtView.ShowBitmap(
  AName : String; ABitmap : TBitmap;
  AIndexTitle : String;
  AIndex : Integer; AAppendIndex : Boolean);
begin
  Bitmap := Nil;
  // Load up the form with the necessary info
  Caption := AName;
  if AAppendIndex then
    SaveDialog.FileName := Format('%s %d.bmp', [StripDamagingChars(AName), AIndex])
  else
    SaveDialog.FileName := StripDamagingChars(AName) + '.bmp';
  lblBlockNumTitle.Caption := AIndexTitle;
  lblBlockNum.Caption := Format('%d (0x%0.4x)', [AIndex, AIndex]);
  lblBlockNum.Left := lblBlockNumTitle.Left + lblBlockNumTitle.Width + 8;
  ClientWidth := Max(ABitmap.Width + 30, 200);
  ClientHeight := pnlHeader.Height + 30 + ABitmap.Height;
  imgTile.SetBounds(
    (ClientWidth-ABitmap.Width-2) div 2, 15,
    ABitmap.Width+2, ABitmap.Height+2);
  with imgTile.Canvas do
    begin
      Pen.Style := psSolid;
      Pen.Color := TColor(DecreaseColorBrightness(ColorToRGB(clWindow), 15));
      Rectangle(0,0,ABitmap.Width+2, ABitmap.Height+2);
      Draw(1,1,ABitmap);
    end;
  // Show the form
  Show;
  // Store a copy of the bitmap last--for saving
  Bitmap := TBitmap.Create;
  Bitmap.Assign(ABitmap);
end;

procedure TfrmArtView.actSaveExecute(Sender: TObject);
var
  Stream : TFileStream;
begin
  Application.ProcessMessages;
  if SaveDialog.Execute and TestFileOverwrite(SaveDialog.FileName) then
    begin
      Stream := TFileStream.Create(SaveDialog.FileName, fmCreate);
      try
        Bitmap.SaveToStream(Stream);
      finally
        Stream.Free;
      end;
    end;
end;

procedure TfrmArtView.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Bitmap.Free;
  Action := caFree;
end;

procedure TfrmArtView.FormResize(Sender: TObject);
begin
  imgTile.Left := (pnlOuter.ClientWidth-imgTile.Width) div 2-1;
  imgTile.Top := (pnlOuter.ClientHeight-imgTile.Height) div 2-1;
end;

end.
