unit TextData;

interface

uses
  Classes;

type
  TTextDataFile = Class (TObject)
  private
    FFileName : String;
    FCurrentSection : String;

    Index : Integer;
    St : TMemoryStream;
  protected
    procedure SetCurrentSection(Value : String);
  public
    constructor Create(AFileName : String);
    destructor Destroy; override;
    procedure ReadCurrentSection(SL : TStringList);

    property FileName : String read FFileName write FFileName;
    property CurrentSection : String read FCurrentSection write SetCurrentSection;
  end;

implementation

uses
  SysUtils;

constructor TTextDataFile.Create(AFileName : String);
begin
  inherited Create;

  FFileName := AFileName;
  St := TMemoryStream.Create;
  try
    St.LoadFromFile(AFileName);
  except
    St.Free;
    raise;
  end;
end;

destructor TTextDataFile.Destroy;
begin
  St.Free;
  inherited Destroy;
end;

procedure TTextDataFile.SetCurrentSection(Value : String);
var
  State, L : Integer;
  C : Char;
  Done : Boolean;
begin
  State := 0;
  L := Length(Value);
  St.Seek(0, soFromBeginning);
  Done := False;
  while (St.Position < St.Size) and not Done do
    begin
      St.Read(C, 1);
      if State = 0 then
        begin if C = '[' then State := 1; end
      else if State <= L then
        begin
          if UpCase(C) = UpCase(Value[State]) then
            Inc(State)
          else
            State := 0;
        end
      else if State = L+1 then
        begin
          if C=']' then
            Inc(State)
          else
            State := 0;
        end
      else if C = #10 then
        Done := True;
    end;

  Index := St.Position;
  FCurrentSection := Value;
end;

function ProcessLine(var Line : String) : Boolean;
var
  St, Ed : Integer;
begin
  Result := False;
  if Length(Line) <= 0 then Exit;

   // Trim the line
  St := 1;
  while Line[St] = ' ' do Inc(St);
  Ed := St;
  while not (Line[Ed] in [';', #00]) do Inc(Ed);
  Dec(Ed);
  while (Ed > 1) and (Line[Ed] = ' ') do Dec(Ed);
  if Ed < St then
    begin
      Line := '';
      Exit;
    end;

  Line := Copy(Line, St, Ed-St+1);

  if Line[1] = '[' then
    begin
      Result := True;
      Exit;
    end;
end;

procedure TTextDataFile.ReadCurrentSection(SL : TStringList);
var
  Line : String;
  Done : Boolean;
  C : Char;
begin
  St.Seek(Index, soFromBeginning);

  Done := False;
  Line := '';
  while not Done do
    begin
      St.Read(C, 1);
      if (C = #10) or (St.Position >= St.Size) then
        begin
          if ProcessLine(Line) then
            Done := True
          else if Line <> '' then
            SL.Add(Line);

          if (St.Position >= St.Size) then
            Done := True;

          Line := '';
        end
      else if C <> #13 then
        Line := Line + C;
    end;
end;

end.
