unit UOAnim;

interface

uses
  Classes, Graphics, ExtCtrls, IndexedFile, UOHues;

type
  TAnimIndexRec = packed record
    Lookup : Integer;
    Length : Integer;
    Unknown : Integer;
  end;

  TAnimFrameHeader = packed record
    ImgCenterX : SmallInt;
    ImgCenterY : SmallInt;
    ImgWidth : SmallInt;
    ImgHeight : SmallInt;
  end;
{
   // Animation classifications
  TCharacterData = class (TObject)
    Handle,
    Index,
    ModelNumber, // Character index including all blanks
    UsageNumber, // Character index excluding all blanks
    ChType : Integer;
  end;

  TActionData = class (TObject)
    Handle,
    Index : Integer;
  end;
}
  TAnimFrame = class (TObject)
    Location : Integer;
    Size : Integer;
    CenterX, CenterY : SmallInt;
    Width, Height : SmallInt;
  end;

  TFrames = Array[0..0] of TAnimFrame;
  PFrames = ^TFrames;

  TFrameSetPalette = Array[0..255] of Word;

  TFrameSet = class (TObject)
  private
    FStream : TMemoryStream;
    FDataStart : Integer;
    Frames : PFrames;
  protected
    function ResetToFrame(FrameNum : Integer) : Integer;
  public
    OriginalPalette,
    Palette : TFrameSetPalette;
    CX : Integer;
    CY : Integer;
    X1, Y1, X2, Y2 : Integer;
    FrameCount : Integer;
    UseHue : Boolean;
    HueEntry : THueEntry;
    TotalHueReplace : Boolean;
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromStream(AStream : TStream; Size : Integer);
    procedure LoadFrameToImage(FrameNum : Integer; Image : TImage; BkColor : TColor; CenterX, CenterY : Integer);
    procedure LoadFrameToBitmap(FrameNum : Integer; Bitmap : TBitmap; CenterX, CenterY : Integer);
    procedure LoadFrameToTextFile(FrameNum : Integer; FileName : String);
    procedure LoadFrameToPointer(FrameNum : Integer; Location : Pointer; CenterX, CenterY, MaxWidth, MaxHeight : Integer);
    procedure LoadRevFrameToPointer(FrameNum : Integer; Location : Pointer; CenterX, CenterY, MaxWidth, MaxHeight : Integer);
  end;

 // Animation
{
function ReworkCharacterLine(Line : String; Data : TCharacterData) : String;
function ReworkActionLine(Line : String; Data : TActionData) : String;
procedure BuildCharacterTypeList(SL : TStringList);
procedure BuildCharacterList(SL : TStringList);
procedure GetCharacterType(ChType : Integer; SL : TStringList);

var
  DatFileName : String = 'InsideUO.dat';
}
implementation

uses
  SysUtils, Utility;

{
function ReworkCharacterLine(Line : String; Data : TCharacterData) : String;
var
  I : Integer;
  C : Char;
  State : Integer;
  Value : String;
  Done : Boolean;
begin
  if Length(Line) <= 0 then Exit;

  State := 0;
  Value := '';
  I := 1;
  Done := False;
  while not Done do
    begin
      C := Line[I];
      if C in [',', #00] then
        case State of
          0 :
            begin
              Data.Handle := StrToInt(Trim(Value));
              State := 1;
              Value := '';
            end;
          1 :
            begin
              Data.ChType := StrToInt(Trim(Value));
              State := 2;
              Value := '';
            end;
          2 :
            begin
              Result := QuTrim(Value);
              Done := True;
            end;
        end
      else
        Value := Value + C;

      Inc(I);
    end;
end;

function ReworkActionLine(Line : String; Data : TActionData) : String;
var
  I : Integer;
  C : Char;
  State : Integer;
  Value : String;
  Done : Boolean;
begin
  if Length(Line) = 0 then Exit;

  State := 0;
  Value := '';
  I := 1;
  Done := False;
  while not Done do
    begin
      C := Line[I];
      case State of
        0 :
          if C = ',' then
            begin
              Data.Handle := StrToInt(Trim(Value));
              State := 1;
              Value := '';
            end
          else if C = #00 then
            raise Exception.Create('Expecting two values, found one')
          else
            Value := Value + C;
        1 :
          if C = #00 then
            begin
              Result := Trim(Value);
              if Result[1] in ['"', ''''] then Delete(Result,1,1);
              if Result[Length(Result)] in ['"', ''''] then Delete(Result,Length(Result),1);
              Done := True;
            end
          else
            Value := Value + C;
      end;

      Inc(I);
    end;
end;

procedure BuildCharacterTypeList(SL : TStringList);
var
  Data : TTextDataFile;
  I : Integer;
begin
   // Load in InsideUO.Dat
  if not FileExists(DatFileName) then
    raise Exception.Create('Cannot find ' + DatFileName);
  Data := TTextDataFile.Create(DatFileName);
  try
    Data.CurrentSection := 'CHARACTER_TYPES';
    Data.ReadCurrentSection(SL);
  finally
    Data.Free;
  end;

   // Convert the string list from the raw format to usable data
  For I := 0 to SL.Count-1 do
    SL[I] := QuTrim(SL[I]);
end;

procedure BuildCharacterList(SL : TStringList);
var
  Data : TTextDataFile;
  ChData : TCharacterData;
  I, J : Integer;
begin

   // Load in InsideUO.Dat
  if not FileExists(DatFileName) then
    raise Exception.Create('Cannot find ' + DatFileName);
  Data := TTextDataFile.Create(DatFileName);
  try
    Data.CurrentSection := 'CHARACTERS';
    Data.ReadCurrentSection(SL);
  finally
    Data.Free;
  end;

   // Convert the string list from the raw format to usable data
  J := 0;
  For I := 0 to SL.Count-1 do
    begin
      ChData := TCharacterData.Create;
      try
        SL[I] := ReworkCharacterLine(SL[I], ChData);
        if SL[I] = '' then
          ChData.UsageNumber := -1
        else
          begin
            ChData.UsageNumber := J;
            Inc(J);
          end;
      except
        ChData.Free;
        Raise;
      end;

      SL.Objects[I] := ChData;
    end;
end;

procedure GetCharacterType(ChType : Integer; SL : TStringList);
var
  Data : TTextDataFile;
  FrData : TActionData;
  I : Integer;
begin
   // Load in InsideUO.Dat
  if not FileExists(DatFileName) then
    raise Exception.Create('Cannot find ' + DatFileName);
  Data := TTextDataFile.Create(DatFileName);
  try
    Data.CurrentSection := 'CHARACTER_TYPE' + IntToStr(ChType);
    Data.ReadCurrentSection(SL);
  finally
    Data.Free;
  end;

   // Convert the string list from the raw format to usable data
  For I := 0 to SL.Count-1 do
    begin
      FrData := TActionData.Create;
      try
        SL[I] := ReworkActionLine(SL[I], FrData);
      except
        FrData.Free;
        Raise;
      end;
      SL.Objects[I] := FrData;
    end;
end;
}
 //-----------------------------------------------------------------------------

constructor TFrameSet.Create;
begin
  inherited Create;
  FStream := TMemoryStream.Create;
end;

destructor TFrameSet.Destroy;
var
  I : Integer;
begin
  If Frames <> Nil then
    begin
      For I := 0 to FrameCount-1 do
        TAnimFrame(Frames^[I]).Free;
      FreeMem(Frames, FrameCount * 4);
    end;

  FStream.Free;
  inherited Destroy;
end;

procedure TFrameSet.LoadFromStream(AStream : TStream; Size : Integer);
type
  TFrameData = Array[0..0] of Integer;
  PFrameData = ^TFrameData;
var
  Frame : TAnimFrame;
  FrameData : PFrameData;
  FrameDataSize : Integer;
  I : Integer;
begin

  FStream.Clear;
  FStream.SetSize(Size);
  FStream.CopyFrom(AStream, Size);
  FStream.Seek(0, soFromBeginning);

   // For Center calculation
  X1 := 1000;
  Y1 := 1000;
  X2 := -1000;
  Y2 := -1000;

  FStream.Read(OriginalPalette, SizeOf(OriginalPalette));
  Palette := OriginalPalette;

  FDataStart := FStream.Position;
  FStream.Read(FrameCount, SizeOf(FrameCount));

  FrameDataSize := FrameCount * 4;
  GetMem(Frames, FrameDataSize);
  FillChar(Frames^, FrameDataSize, 0);

  GetMem(FrameData, FrameCount * 4);
  try
    FStream.Read(FrameData^, FrameDataSize);

    For I := 0 to FrameCount-1 do
      begin
        Frame := TAnimFrame.Create;
        Frames^[I] := Frame;
        Frame.Location := FrameData^[I];
        if I < FrameCount-1 then
          Frame.Size := FrameData^[I+1] - FrameData^[I]
        else
          Frame.Size := Size - FrameData^[I];

        if Frame.Size < 0 then
          Frame.Size := 0;

        FStream.Seek(FDataStart + Frame.Location, soFromBeginning);
        FStream.Read(Frame.CenterX, SizeOf(Frame.CenterX));
        FStream.Read(Frame.CenterY, SizeOf(Frame.CenterY));
        FStream.Read(Frame.Width, SizeOf(Frame.Width));
        FStream.Read(Frame.Height, SizeOf(Frame.Height));

        with Frame do
          begin
            if -CenterX < X1 then  X1 := -CenterX;
            if -CenterX + Width > X2 then X2 := -CenterX + Width;
            if -CenterY - Height < Y1 then Y1 := -CenterY - Height;
            if -CenterY > Y2 then Y2 := -CenterY;
          end;
      end;
  finally
    FreeMem(FrameData, FrameDataSize);
  end;

  CX := (X2 + X1) div 2;
  CY := (Y2 + Y1) div 2;
end;

function TFrameSet.ResetToFrame(FrameNum : Integer) : Integer;
var
  Location : Integer;
begin
  if (FrameNum < 0) or (FrameNum >= FrameCount) then
    raise Exception.Create('FrameNum out of range in FrameSet');
  Location := FDataStart + Frames^[FrameNum].Location;
  Result := Location + Frames^[FrameNum].Size;
  FStream.Seek(Location, soFromBeginning);
end;

procedure TFrameSet.LoadFrameToBitmap(FrameNum : Integer; Bitmap : TBitmap; CenterX, CenterY : Integer);
 // Displays a frame from the frameset in a Bitmap.
var
  EndOfData : Integer;
  PrevLineNum, LineNum : Byte;
  Unknown : ShortInt;
  RunLength, RowHeader, RowOfs : SmallInt;
  StX, StY, Y : Integer;
  Row : PWordArray;
  MaxWidth, MaxHeight : Integer;
  Header : TAnimFrameHeader;
  DataLoc : PChar;

  I : Integer;
  C : Word;
begin
  Palette := OriginalPalette;
  if UseHue then
    begin
      for I := 0 to 255 do
        begin
          C := Palette[I];
          // If it is a total replace, then convert every pixel to grayscale
          // and map to the hue lookup.
          if TotalHueReplace then
            Palette[I] := HueEntry.ColorTable[ConvertToGray(C)]
          // Otherwise we just replace pixels that are already gray.
          else if (C and $1F = (C shr 5) and $1F) and
                  (C and $1F = (C shr 10) and $1F) then
                Palette[I] := HueEntry.ColorTable[C and $1F];
        end;
    end;

  EndOfData := ResetToFrame(FrameNum);
  FStream.Read(Header, SizeOf(Header));

  MaxWidth := Bitmap.Width;
  MaxHeight := Bitmap.Height;

  StY := CenterY - Header.ImgCenterY - Header.ImgHeight;
  //StX := CenterX - Header.ImgCenterX;

  {
  if cbAnimRect.Checked then
    with imgAnimation.Canvas do
      begin
        Pen.Style := psSolid;
        Pen.Color := clRed;
        Rectangle(StX, StY, StX + ImgWidth, StY + ImgHeight);
      end;
  }

  PrevLineNum := $FF;
  Y := StY;
  while FStream.Position < EndOfData do
    begin
       // Load row header
      FStream.Read(RowHeader, 2);
      FStream.Read(RowOfs, 2);

      if (RowHeader = $7FFF) or (RowOfs = $7FFF) then Exit;
      asm
        push eax
        push ebx

         // Extract RunLength
        mov  ax, RowHeader
        mov  bx, ax
        and  bx, $0FFF
        mov  RunLength, bx

         // Exctract LineNum
        shr  ax, 12
        mov  LineNum, al

         // Extract Unknown
        mov  ax, RowOfs
        mov  bx, ax
        and  bx, $3F
        mov  Unknown, bl

         // Recalculate RowOfs
        sar  ax, 6
        mov  RowOfs, ax

         // Calculate StX = CenterX + RowOfs
        movsx eax, ax
        add  eax, CenterX
        mov  StX, eax

         // if (PrevLineNum <> $FF) and (LineNum <> PrevLineNum) then Inc(Y);
         // PrevLineNum := LineNum;
        mov  al, LineNum
        cmp  PrevLineNum, $FF
        je   @NotNextLine
        cmp  al, PrevLineNum
        je   @NotNextLine
        inc  Y
       @NotNextLine:
        mov  PrevLineNum, al

        pop  ebx
        pop  eax
      end;

      if Y >= 0 then
        begin
          if Y >= MaxHeight then Exit;

           // Draw row
          Row := Bitmap.ScanLine[Y];
          DataLoc := PChar(FStream.Memory)+FStream.Position;

          asm
            push edi
            push esi
            push ebx

             // ebx = Palette
            mov  ebx, Self
            add  ebx, Palette

             // eax = Positive(StX)
             // edx = Negative(StX)
            xor  eax, eax
            mov  edx, StX
            cmp  edx, 0
            jl @LeftClip
            mov  eax, edx
            mov  edx, 0
           @LeftClip:

             // ecx = Counter
            movsx ecx, RunLength
            add  ecx, edx
            cmp  ecx, 0    // May not be necessary
            jle  @Done

             // esi = Source (DataLoc)
            mov  esi, DataLoc
            sub  esi, edx

             // edi = Dest (Row)
             // edx = Max value of edi
            mov  edi, Row
            mov  edx, MaxWidth
            shl  edx, 1
            add  edx, edi
            shl  eax, 1
            add  edi, eax

             // process a row
           @StartOfRow:
            cmp  edi, edx    // Boundary check on the right
            jg  @Done
            xor  eax,eax
            mov  al, [esi]   // Read byte from source
            inc  esi
            shl  eax, 1
            add  eax, ebx
            mov  ax, [eax]   // Translate to word through Palette
            stosw            // Store color value
            loop @StartOfRow

           @Done:
            pop  ebx
            pop  esi
            pop  edi
          end;
        end;

      FStream.Seek(RunLength, soFromCurrent);
    end;
end;  // LoadFrameToBitmap

procedure TFrameSet.LoadFrameToTextFile(FrameNum : Integer; FileName : String);
var
  EndOfData : Integer;
  F : Text;
  LineNum : Byte;
  Unknown : ShortInt;
  RunLength, RowHeader, RowHeader2 : SmallInt;
  RowOfs : SmallInt;
  Header : TAnimFrameHeader;
  I : Integer;
  B : Byte;
begin
  EndOfData := ResetToFrame(FrameNum);

  AssignFile(F, FileName);
  Rewrite(F);
  try
    FStream.Read(Header, SizeOf(Header));

    WriteLn(F, 'CenterX = ' + IntToStr(Header.ImgCenterX));
    WriteLn(F, 'CenterY = ' + IntToStr(Header.ImgCenterY));
    WriteLn(F, 'Width = ' + IntToStr(Header.ImgWidth));
    WriteLn(F, 'Height = ' + IntToStr(Header.ImgHeight));
    WriteLn(F);

    while FStream.Position < EndOfData do
      begin
         // Load row header
        FStream.Read(RowHeader, 2);
        FStream.Read(RowHeader2, 2);

        if (RowHeader = $7FFF) or (RowOfs = $7FFF) then Exit;

        asm
          push eax
          push ebx

           //  LLLL RRRR RRRR RRRR
           //  0000 0000 0000 0000

           // Extract RunLength
          mov  ax, RowHeader
          mov  bx, ax
          and  bx, $0FFF
          mov  RunLength, bx

           // Exctract LineNum
          shr  ax, 12
          mov  LineNum, al

           //  RRRR RRRR RRUU UUUU
           //  0000 0000 0000 0000

           // Extract Unknown
          mov  ax, RowHeader2
          mov  bx, ax
          and  bx, $3F
          mov  Unknown, bl

           // Recalculate RowOfs
          sar  ax, 6
          mov  RowOfs, ax

          pop  ebx
          pop  eax
        end;

        Write(F, Format('%0.2x %0.2x %d %d ', [LineNum, Unknown, RunLength, RowOfs]));

         // Recalc and verify Y
        //if (PrevLineNum <> $FF) and (LineNum <> PrevLineNum) then Inc(Y);
        //PrevLineNum := LineNum;
        For I := 0 to RunLength-1 do
          begin
            FStream.Read(B, 1);
            Write(F, Format('%0.2x ', [B])); // Palette[B]
          end;

        WriteLn(F);
      end;
  finally
    CloseFile(F);
  end;
end;  // LoadFrameToTextFile

procedure TFrameSet.LoadFrameToImage(FrameNum : Integer; Image : TImage; BkColor : TColor; CenterX, CenterY : Integer);
 // Displays a frame from the frameset in a TImage (first flushing the
 // background to BKColor).
begin
  with Image.Picture.Bitmap do
    begin
      Width := Image.Width;
      Height := Image.Height;
      PixelFormat := pf15bit;
      HandleType := bmDIB;
    end;

  with Image.Picture.Bitmap.Canvas do
    begin
      Brush.Color := BkColor;
      FillRect(Image.ClientRect);
    end;

  LoadFrameToBitmap(FrameNum, Image.Picture.Bitmap,
                    CenterX, CenterY);
end;  // LoadFrameNumToImage

procedure TFrameSet.LoadFrameToPointer(FrameNum : Integer; Location : Pointer; CenterX, CenterY, MaxWidth, MaxHeight : Integer);
 // Displays a frame from the frameset in a Bitmap.
var
  EndOfData : Integer;
  PrevLineNum, LineNum : Byte;
  Unknown : ShortInt;
  RunLength, RowHeader, RowOfs : SmallInt;
  StX, Y : Integer;
  Header : TAnimFrameHeader;
  DataLoc : PChar;
begin
  EndOfData := ResetToFrame(FrameNum)-FStream.Position;
  DataLoc := PChar(FStream.Memory)+FStream.Position;
  asm
    pushad
    cld

     // ebx = Palette
    mov  ebx, Self
    add  ebx, Palette

     // Load Header
    mov  esi, DataLoc
    lea  edi, Header

     // Recalc EndOfData
    mov  edx, EndOfData
    add  edx, esi
    mov  EndOfData, edx

     // Finish loading header
    mov  ecx, 8
    rep  movsb

     // Calc Y
    mov  eax, CenterY
    movsx ecx, Header.ImgCenterY
    sub  eax, ecx
    movsx ecx, Header.ImgHeight
    sub  eax, ecx
    mov  Y, eax

     // while Position < EndOfData do
    mov  PrevLineNum, $FF
   @ProcessRow:
    cmp  esi, EndOfData
    jge @Done

       // Load RowHeader and RowOfs
      mov  ax, [esi]
      cmp  ax, $7FFF
      je  @Done
      mov  RowHeader, ax

      mov  cx, [esi+2]
      cmp  cx, $7FFF
      je  @Done
      mov  RowOfs, cx
      add  esi, 4

       // Extract RunLength
      mov  dx, ax
      and  dx, $0FFF
      mov  RunLength, dx

       // Exctract LineNum
      shr  ax, 12
      mov  LineNum, al

       // Extract Unknown
      mov  dx, cx
      and  dx, $3F
      mov  Unknown, dl

       // Recalculate RowOfs
      sar  cx, 6
      mov  RowOfs, cx

       // Calculate StX = CenterX + RowOfs
      movsx eax, cx
      add  eax, CenterX
      mov  StX, eax

       // if (PrevLineNum <> $FF) and (LineNum <> PrevLineNum) then Inc(Y);
       // PrevLineNum := LineNum;
      mov  al, LineNum
      cmp  PrevLineNum, $FF
      je   @NotNextLine
      cmp  al, PrevLineNum
      je   @NotNextLine
      inc  Y
     @NotNextLine:
      mov  PrevLineNum, al

      mov  eax, 640 * 2
      mul  Y
      add  eax, Location
      mov  edi, eax

       // Vertically bounds check row
      mov  eax, Y
      cmp  eax, 0
      jl  @DontDrawRow
      cmp  eax, MaxHeight
      jge @Done

       // eax = Positive(StX)
       // edx = Negative(StX)
      xor  eax, eax
      mov  edx, StX
      cmp  edx, 0
      jl @LeftClip
      mov  eax, edx
      mov  edx, 0
     @LeftClip:

       // ecx = Counter
      movsx ecx, RunLength
      add  ecx, edx
      cmp  ecx, 0    // May not be necessary
      jle  @DontDrawRow

       // esi = Source (DataLoc)
      sub  esi, edx

       // edi = Dest (Row)
       // edx = Max value of edi
      mov  edx, MaxWidth
      shl  edx, 1
      add  edx, edi
      shl  eax, 1
      add  edi, eax

       // process a row
     @StartOfRow:
      cmp  edi, edx    // Boundary check on the right
      jge  @RowDone
      xor  eax,eax
      mov  al, [esi]   // Read byte from source
      inc  esi
      shl  eax, 1
      add  eax, ebx
      mov  ax, [eax]   // Translate to word through Palette
      stosw            // Store color value
      loop @StartOfRow

     @RowDone:
      add  esi, ecx    // Jump to the EOL
      jmp @ProcessRow

     @DontDrawRow:
      movsx eax, RunLength
      add  esi, eax
      jmp @ProcessRow
   @Done:
    popad
  end;
end;  // LoadFrameToBitmap

procedure TFrameSet.LoadRevFrameToPointer(FrameNum : Integer; Location : Pointer; CenterX, CenterY, MaxWidth, MaxHeight : Integer);
 // Displays a frame from the frameset in a Bitmap.
var
  EndOfData : Integer;
  PrevLineNum, LineNum : Byte;
  Unknown : ShortInt;
  RunLength, RowHeader, RowOfs : SmallInt;
  StX, Y : Integer;
  Header : TAnimFrameHeader;
  DataLoc : PChar;
begin
  EndOfData := ResetToFrame(FrameNum)-FStream.Position;
  DataLoc := PChar(FStream.Memory)+FStream.Position;
  asm
    pushad
    cld

     // ebx = Palette
    mov  ebx, Self
    add  ebx, Palette

     // Load Header
    mov  esi, DataLoc
    lea  edi, Header

     // Recalc EndOfData
    mov  edx, EndOfData
    add  edx, esi
    mov  EndOfData, edx

     // Finish loading header
    mov  ecx, 8
    rep  movsb

     // Calc Y
    mov  eax, CenterY
    movsx ecx, Header.ImgCenterY
    sub  eax, ecx
    movsx ecx, Header.ImgHeight
    sub  eax, ecx
    mov  Y, eax

     // while Position < EndOfData do
    mov  PrevLineNum, $FF
   @ProcessRow:
    cmp  esi, EndOfData
    jge @Done

       // Load RowHeader and RowOfs
      mov  ax, [esi]
      cmp  ax, $7FFF
      je  @Done
      mov  RowHeader, ax

      mov  cx, [esi+2]
      cmp  cx, $7FFF
      je  @Done
      mov  RowOfs, cx
      add  esi, 4

       // Extract RunLength
      mov  dx, ax
      and  dx, $0FFF
      mov  RunLength, dx

       // Exctract LineNum
      shr  ax, 12
      mov  LineNum, al

       // Extract Unknown
      mov  dx, cx
      and  dx, $3F
      mov  Unknown, dl

       // Recalculate RowOfs
      sar  cx, 6
      mov  RowOfs, cx

       // Calculate StX = CenterX - RowOfs - RunLength
      movsx eax, cx
      movsx edx, RunLength
      neg  eax
      sub  eax, edx
      add  eax, CenterX
      mov  StX, eax

       // if (PrevLineNum <> $FF) and (LineNum <> PrevLineNum) then Inc(Y);
       // PrevLineNum := LineNum;
      mov  al, LineNum
      cmp  PrevLineNum, $FF
      je   @NotNextLine
      cmp  al, PrevLineNum
      je   @NotNextLine
      inc  Y
     @NotNextLine:
      mov  PrevLineNum, al

      mov  eax, 640 * 2
      mul  Y
      add  eax, Location
      mov  edi, eax

       // Shift the source over to the end
      movsx ecx, RunLength
      add  esi, ecx
      dec  esi

       // Vertically bounds check row
      mov  eax, Y
      cmp  eax, 0
      jl  @DontDrawRow
      cmp  eax, MaxHeight
      jge @Done

       // eax = Positive(StX)
       // edx = Negative(StX)
      xor  eax, eax
      mov  edx, StX
      cmp  edx, 0
      jl @LeftClip
      mov  eax, edx
      mov  edx, 0
     @LeftClip:

       // ecx = Counter
      add  ecx, edx
      cmp  ecx, 0    // May not be necessary
      jle  @DontDrawRow

       // esi = Source (DataLoc)
      add  esi, edx

       // edi = Dest (Row)
       // edx = Max value of edi
      mov  edx, MaxWidth
      shl  edx, 1
      add  edx, edi
      shl  eax, 1
      add  edi, eax

       // process a row
     @StartOfRow:
      cmp  edi, edx    // Boundary check on the right
      jge  @RowDone
      xor  eax,eax
      mov  al, [esi]   // Read byte from source
      dec  esi
      shl  eax, 1
      add  eax, ebx
      mov  ax, [eax]   // Translate to word through Palette
      stosw            // Store color value
      loop @StartOfRow

     @RowDone:
      sub  esi, ecx    // Jump to the EOL
      movsx eax, RunLength
      add  esi, eax
      inc  esi
      jmp @ProcessRow

     @DontDrawRow:
      inc  esi   // It's already pointing to the last pixel
      jmp @ProcessRow
   @Done:
    popad
  end;
end;  // LoadRevFrameToBitmap

end.
