unit UOMulti;

interface

uses
  Classes, Graphics, SysUtils, UOArt;

type
{$A-}
  TMultiIndex = Record
    Lookup,
    Length,
    Unknown : Integer;
  end;
{$A+}

procedure DrawMultiToBitmap(
  Stream : TStream;
  ArtCacheIndex : TArtCacheIndex;
  ShowFloors : Integer;
  Bitmap : TBitmap);

implementation

type
  TMulti = class (TObject)
    BlockNum : SmallInt;
    X : SmallInt;
    Y : SmallInt;
    Alt : SmallInt;
    Flags : Integer;
    Width,
    Height : Word;
    Style : TBlockStyle;
  end;

procedure DrawMultiToBitmap(
  Stream : TStream;
  ArtCacheIndex : TArtCacheIndex;
  ShowFloors : Integer;
  Bitmap : TBitmap);
var
  X1, Y1, X2, Y2, TmpX, TmpY, I, CX, CY, AX, AY : Integer;
  ObjList : TList;
  Multi : TMulti;
begin
  X1 := 1000; Y1 := 1000; X2 := -1000; Y2 := -1000;

  ObjList := TList.Create;
  try
     // Break the multi structure down into TMulti objects.
    while Stream.Position < Stream.Size do
      begin
        Multi := TMulti.Create;
        try
          Stream.Read(Multi.BlockNum, SizeOf(Multi.BlockNum));
          ArtCacheIndex.GetBlockInfo(16384 + Multi.BlockNum, Multi.Width, Multi.Height, Multi.Style);
          Stream.Read(Multi.X, SizeOf(Multi.X));
          Stream.Read(Multi.Y, SizeOf(Multi.Y));
          Stream.Read(Multi.Alt, SizeOf(Multi.Alt));
          Stream.Read(Multi.Flags, SizeOf(Multi.Flags));

          TmpX := (Multi.X - Multi.Y) * 22 - (Multi.Width shr 1);
          TmpY := (Multi.X + Multi.Y) * 22 - Multi.Height - Multi.Alt * 4;

          if TmpX < X1 then X1 := TmpX;
          if TmpY < Y1 then Y1 := TmpY;
          if TmpX+Multi.Width > X2 then X2 := TmpX+Multi.Width;
          if TmpY+Multi.Height > Y2 then Y2 := TmpY+Multi.Height;
        except
          Multi.Free;
          raise;
        end;

        I := 0;
         // Sort by row
        while (I < ObjList.Count) and
              (Multi.X+Multi.Y > TMulti(ObjList[I]).X+TMulti(ObjList[I]).Y) do
          Inc(I);
         // Then by altitude
        while (I < ObjList.Count) and
              (Multi.X+Multi.Y = TMulti(ObjList[I]).X+TMulti(ObjList[I]).Y) and
              (Multi.Alt > TMulti(ObjList[I]).Alt) do
          Inc(I);
         // Then by height
         // TODO: This should actually be the actual altitude height rather than
         // the number of pixels high (since tiles are frequently taller than
         // short blocks.  However, that height looks like it is the last byte
         // in TileData which is not currently cached.
        while (I < ObjList.Count) and
              (Multi.X+Multi.Y = TMulti(ObjList[I]).X+TMulti(ObjList[I]).Y) and
              (Multi.Alt = TMulti(ObjList[I]).Alt) and
              (Multi.Height >= TMulti(ObjList[I]).Height) do
          Inc(I);

        ObjList.Insert(I, Multi);
      end;
    CX := -X1;
    CY := -Y1;

    // Establish the image based on the size of the Multi
    with Bitmap do
      begin
        Transparent := False;
        PixelFormat := pf15bit;
        Width := X2-X1;
        Height := Y2-Y1;
        Canvas.Brush.Color := clWindow;
        Canvas.FillRect(Rect(0,0,Width,Height));
      end;

     // Draw the multi
    for I := 0 to ObjList.Count-1 do
      with TMulti(ObjList[I]) do
        if Alt < ShowFloors * 20 then
          begin
            AX := CX + (X - Y) * 22 - (Width shr 1);
            AY := CY + (X + Y) * 22 - Height - Alt * 4;
            ArtCacheIndex.DrawBlock(16384 + BlockNum, Bitmap, AX, AY);
          end;

  finally
    for I := 0 to ObjList.Count-1 do
      TMulti(Objlist[I]).Free;
    ObjList.Free;
  end;
end;

end.
