unit Utility;

interface

function Min(I1, I2 : Integer) : Integer;
function Max(I1, I2 : Integer) : Integer;
function QuTrim(S : String) : String;
function DecreaseColorBrightness(Color : LongInt; Amount: Byte) : LongInt;
function TestFileOverwrite(FileName : String) : Boolean;

implementation

uses
  Controls, Dialogs, SysUtils;

const
  VerifyFileOverwrite =
    'Are you sure you want to overwrite the file %s?';

function Min(I1, I2 : Integer) : Integer;
begin
  if I1 < I2 then Result := I1 else Result := I2;
end;

function Max(I1, I2 : Integer) : Integer;
begin
  If I1 > I2 then Result := I1 else Result := I2;
end;

function QuTrim(S : String) : String;
begin
  Result := Trim(S);
  if Result[1] in ['"', ''''] then Delete(Result,1,1);
  if Result[Length(Result)] in ['"', ''''] then Delete(Result,Length(Result),1);
end;

function DecreaseColorBrightness(Color : LongInt; Amount: Byte) : LongInt;
var
  Red, Green, Blue : Byte;
begin
  Red := (Color shr 16) and $FF;
  Green := (Color shr 8) and $FF;
  Blue := Color and $FF;
  if(Red >= Amount) then
    Dec(Red,Amount)
  else
    Red := 0;
  if(Green >= Amount) then
    Dec(Green,Amount)
  else
    Green := 0;
  if(Blue >= Amount) then
    Dec(Blue,Amount)
  else
    Blue := 0;
  Result := (Red shl 16) + (Green shl 8) + Blue;
end;

// Tests to see if an overwrite message needs to be displayed for a file
// and displays it if it does.  If the function returns true, then you should
// go ahead and overwrite the file.  If it returns false, you shouldn't.
function TestFileOverwrite(FileName : String) : Boolean;
begin
  if FileExists(FileName) then
    begin
      Result := MessageDlg(
        Format(VerifyFileOverwrite, [FileName]),
        mtConfirmation, mbOKCancel, 0) = mrOK;
    end
  else
    Result := true;
end;

end.
